/**
 * A sample lever to be triggered by the Use() exec function.
 * Triggers have to be used in Kismet, otherwise Use() will not find it
 * 
 * 2013 by FreetimeCoder
 * www.freetimestudio.net
 */
class AdvKitLeverTrigger extends Trigger
    placeable
	ClassGroup(AdvKit);

//mesh to represent the lever
var     SkeletalMeshComponent   mesh;
//node to control the lever's animations
var     AnimNodePlayCustomAnim  customAnim;

//simple switch variable to store the levers state
var()   bool    bActivated;

//local offset to set the Pawn to that is using the lever
var()   Vector  vPawnOffset;

//animations for the lever
var     name    nAnim_Activation;
var     name    nAnim_Deactivation;

/**Actor.uc: 
 * called after initializing the AnimTree for the given SkeletalMeshComponent that has this Actor as its Owner
 * this is a good place to cache references to skeletal controllers, etc that the Actor modifies
 * 
 * @param   _skelComp   the SkeletalMeshComponent that has been initialized
 */
simulated event PostInitAnimTree(SkeletalMeshComponent _skelComp)
{
	super.PostInitAnimTree(_skelComp);

	//find the CustomAnim node within the Anim Tree
	if (_skelComp == mesh)
	{
		customAnim = AnimNodePlayCustomAnim(_skelComp.FindAnimNode('CustomAnim'));
	}
}

/**Actor.uc:
 * Event called when an AnimNodeSequence (in the animation tree of one of this Actor's SkeletalMeshComponents) reaches the end and stops.
 * Will not get called if bLooping is 'true' on the AnimNodeSequence.
 * bCauseActorAnimEnd must be set 'true' on the AnimNodeSequence for this event to get generated.
 *
 * @param	SeqNode		- Node that finished playing. You can get to the SkeletalMeshComponent by looking at SeqNode->SkelComponent
 * @param	PlayedTime	- Time played on this animation. (play rate independant).
 * @param	ExcessTime	- how much time overlapped beyond end of animation. (play rate independant).
 */
event OnAnimEnd(AnimNodeSequence SeqNode, float PlayedTime, float ExcessTime)
{
	//Switch state when lever is finished toggle animation
	switch(SeqNode.AnimSeqName)
	{
	case 'Activate':
	case 'Deactivate':
		bActivated = !bActivated;
		break;
	}
}

/**Actor.uc:
 * "Called when being activated by the specified pawn.  Default
 * implementation searches for any SeqEvent_Used and activates
 * them."
 * 
 * This has been modified to animate the lever and check if the 
 * pawn is of the type AdvKitPawn
 *
 * @return		true to indicate this actor was activated
 */
function bool UsedBy(Pawn User)
{
	local Vector     vOffsetLocation;
	local AdvKitPawn advPawn;

	advPawn = AdvKitPawn(User);
	if(advPawn != none)
    {
		//play animation depending on state
		if(!bActivated)
		{
			if(advPawn.Controller.IsA('AdvKitPlayerController'))
			{
				AdvKitPlayerController(advPawn.Controller).Interact(EIS_ActivateLever);
			}
			if(customAnim!=none)
			{
				customAnim.PlayCustomAnim(nAnim_Activation,1,0,0,false);
			}
		}
		else
		{
			if(advPawn.Controller.IsA('AdvKitPlayerController'))
			{
				AdvKitPlayerController(advPawn.Controller).Interact(EIS_DeactivateLever);
			}
			if(customAnim!=none)
			{
				customAnim.PlayCustomAnim(nAnim_Deactivation,1,0,0,false);
			}
		}

		//position the pawn to sync the animations
		vOffsetLocation = vPawnOffset >> Rotation;
		User.ZeroMovementVariables();
		advPawn.SetLocation(Location+vOffsetLocation);
		advPawn.SetRotation(Rotation);
		
		//notify Kismet etc.
		return super.UsedBy(User);
    }

	//done nothing
    return false;
}

defaultproperties
{
	//do not show the sprite while in game
	Begin Object Name=Sprite
		HiddenGame=true
	End Object

	//set up the mesh to show in game
	begin object Class=SkeletalMeshComponent Name=MeshComp0
		AnimSets(0)=AnimSet'AdvKit_Assets.Lever.S_Lever_Anims'
		AnimTreeTemplate=AnimTree'AdvKit_Assets.Lever.AT_Lever'
		SkeletalMesh=SkeletalMesh'AdvKit_Assets.Lever.S_Lever'

		//I messed up the mesh transform, so I have to correct it here
		Scale = 13
		Translation=(Z=-40)

		CastShadow=true
		bCastDynamicShadow=true
	End Object
	mesh=MeshComp0;
	Components.Add(MeshComp0)

	//show in game
	bHidden = false

	vPawnOffset = (X=-28.0,Y=0.0,Z=15.0)

	//assign animation names
	nAnim_Activation = Activate
	nAnim_Deactivation = Deactivate
}