/**
 * Adventure Kit Player Controller
 * 
 * 2013 by FreetimeCoder
 * www.freetimestudio.net
 */
class AdvKitPlayerController extends PlayerController;

/**hold a AdvKitPawn reference to prevent alot of unecessary casting (set in Posses)*/
var AdvKitPawn advPawn;

//NOTE: since FOV is set in the PlayerController, I figured I'd place
//the camera distance variables here as well
/**minimum distance of the camera to the pawn*/
var float fMinCameraDistance;
/**maximum distance of the camera to the pawn*/
var float fMaxCameraDistance;

/*----------------------------------------------------------------------------------------*\
|*------------------"Hotfix" Utility Functions for easy testing---------------------------*|
\*----------------------------------------------------------------------------------------*/

//These functions are easy wrappers to make the AdvKit work out of the box. You should
//replace these with your own implementations

//This is just to make pressing ESC quit the game for easier testing
exec function ShowMenu()
{
	ConsoleCommand("quit");
}

//Wrapper for zooming in
exec function NextWeapon() 
{
	Zoom(64,1);
}

//Wrapper for zooming out
exec function PrevWeapon()
{
	Zoom(-64,1);
}


/*----------------------------------------------------------------------------------------*\
|*-------------------------------------Input handling-------------------------------------*|
\*----------------------------------------------------------------------------------------*/

/**This calculates a relative movement input vector with the axes being within -1 and 1
 * (Movement input is keys or left analog stick)
 * 
 * @param   _fDeltaTime the current delta time to use for scaling
 * @return              the relative vector
 */
final function Vector GetRelativeInputMovement(float _fDeltaTime)
{
	local Vector vInputVector;

	vInputVector.X = PlayerInput.aForward/PlayerInput.MoveForwardSpeed;
	vInputVector.Y = PlayerInput.aStrafe/PlayerInput.MoveStrafeSpeed;
	vInputVector.Z = PlayerInput.aUp/PlayerInput.MoveStrafeSpeed;

	if(_fDeltaTime!=0)
	{
		vInputVector /= 100.f*_fDeltaTime;
	}

	//NOTE: For some reason the Z Axis scaling differs from XY, so it needs to be clamped
	vInputVector.Z = FClamp(vInputVector.Z,-1,1);

	return vInputVector;
}

/**This calculates a relative look input vector with the axes being within -1 and 1
 * (Look input is mouse or right analog stick)
 * 
 * @param   _fDeltaTime	the current delta time to use for scaling
 * @return              the relative vector
 */
final function Vector GetRelativeInputLook(float _fDeltaTime)
{
	local Vector vInputVector;

	vInputVector.X = PlayerInput.aTurn/PlayerInput.LookRightScale;
	vInputVector.Y = PlayerInput.aLookUp/PlayerInput.LookUpScale;
	vInputVector.Z = 0; //Nothing

	if(_fDeltaTime!=0)
	{
		vInputVector /= 100.f*_fDeltaTime;
	}

	return vInputVector;
}

/**Transforms relative movement input into camera aligned axes
 * For walking this means the XY-Plane
 * 
 * @param   _fDeltaTime	the current delta time to use for scaling
 * @return              the transformed vector
 */
function Vector GetDirectionalizedInputMovement(float _fDeltaTime)
{
	local Vector	X,Y,Z, vInputVector;

	vInputVector = GetRelativeInputMovement(_fDeltaTime);
	vInputVector.Z = 0;

	//Prevent speed hack (e.g. Forward is slower than Foward + Right movement)
	if(VSize(vInputVector)>1)
		vInputVector = Normal(vInputVector);

	//Align acceleration to view angle
	GetAxes(PlayerCamera.Rotation,X,Y,Z);

	X.Z = 0;
	Y.Z = 0;

	vInputVector = Normal(vInputVector.X*X + vInputVector.Y*Y);

	return vInputVector;
}





/**Check if the pawn is standing in front of a platform that he can vault on
 * 
 * @return  true in case the Pawn can vault, false otherwise
 */
final function bool CheckVaulting()
{
	local Vector vPawnForwardAxis, vPawnSideAxis;
	local Vector vTraceStart, vTraceDirection;
	local Vector vLastTraceStartNear, vLastTraceStartFar;

	vPawnForwardAxis = vect(1,0,0) >> Pawn.Rotation;

	vTraceStart = advPawn.Location + vect(0,0,1)*advPawn.fVaultHeight + vPawnForwardAxis*advPawn.fVaultLookAheadDist;
	vTraceDirection = - vect(0,0,1)*advPawn.fVaultTraceDepth;

	//Vertical Cast (Is there a floor?)
	if(!FastTrace(vTraceStart+vTraceDirection,vTraceStart))
	{
		vTraceStart = advPawn.Location + vect(0,0,1)*advPawn.fVaultHeight;
		vTraceDirection = vPawnForwardAxis*(advPawn.fVaultTraceDepth+advPawn.fVaultLookAheadDist);

		//Horizontal Cast (Is there a wall in the way?)
		if(FastTrace(vTraceStart+vTraceDirection,vTraceStart))
		{
			//There is no wall, this is a platform
			//Make sure there is enough space on it
			vPawnSideAxis = vect(0,1,0) >> Pawn.Rotation;
			vLastTraceStartNear = vTraceStart - vPawnSideAxis*Pawn.GetCollisionRadius();
			vLastTraceStartFar = vLastTraceStartNear + vTraceDirection;

			vTraceDirection = vPawnSideAxis*Pawn.GetCollisionRadius()*2;

			if(FastTrace(vLastTraceStartNear+vTraceDirection,vLastTraceStartNear) && FastTrace(vLastTraceStartFar+vTraceDirection,vLastTraceStartFar))
			{
				return true;
			}
		}
	}

	return false;
}

/**Check if the Pawn is on a slide
 * I case it is, this function sets the Slope-variables of the pawn
 * 
 * @return  true if it is on a slide
 */
final function bool CheckIfOnSlide()
{
	local Vector vTraceStart, vTraceEnd, vHitNormal;
	local AdvKitPhysicalMaterialProperty matProperty;

	//trace straight down
	vTraceStart = advPawn.Location;
	vTraceEnd = vTraceStart + vect(0,0,-1) * advPawn.fSlideTraceLength;

	if(TraceAdvanced(vTraceStart,vTraceEnd,,,,vHitNormal,,,matProperty) != none)
	{
		//if the sufrace is a slide, it has to have a pysical material
		if(matProperty!=none && matProperty.bIsSlide)
		{
			//set Pawn's variables from slide
			advPawn.fSlideControlStrength = matProperty.fSlideControlStrength;
			advPawn.vSlopeNormal = vHitNormal;
			advPawn.vSlopeTangent = vHitNormal-vect(0,0,1);
			return true;
		}
	}

	//no slide
	return false;
}







/*----------------------------------------------------------------------------------------*\
|*-------------------------------------Spline Movment-------------------------------------*|
\*----------------------------------------------------------------------------------------*/

/**Finds the closest location on the closest spline actor to a given position within a given radius
 * 
 * @param   _vToLocation    the reference location to
 * @param   _splineClass    only spline actors with this class will be chacked
 * @param   _outSpline      the spline actor closest to the reference location
 * @param   _outfLength     the length on the spline closest to the reference location
 * @param   _fCheckRadius   how far around the location to check for spline actors
 * @param   _fAcceptRadius  the threshhold to accept locations on spline actors
 * @return                  true if a spline actor was found, false if not
 */
final function bool GetClosestSpline(Vector _vToLocation, class<AdvKitSplineActor> _splineClass, out AdvKitSplineActor _outSpline, optional out float _outfLength, optional float _fCheckRadius, optional float _fAcceptRadius)
{
	local Actor tmpActor;
	local AdvKitSplineActor closestSpline, tempSpline, checkSpline;
	local Vector vTempLocation;
	local float fClosestDist, fClosestLength;
	local float fTempLength, fTempDist;

	//iterate through all spliness
	foreach DynamicActors(_splineClass,tmpActor)
	{
		tempSpline = AdvKitSplineActor(tmpActor);

		//if check radius is 0 check all splines, otherwise check only those within the radius
		if(_fCheckRadius<=0 ||  VSize(tempSpline.Location - _vToLocation) <= _fCheckRadius)
		{
			checkSpline = none;
			//the spline actor needs to have a connection
			if(tempSpline.HasNext())
			{
				checkSpline = tempSpline;
			}
			//otherwhise look if the found actor is the end of another one
			else if(tempSpline.HasPrevious())
			{
				checkSpline = tempSpline.GetPrevious();
			}

			if(checkSpline!=none)
			{
				//get closest location on this spline
				fTempDist = checkSpline.GetClosestPositionOnSplineTo(_vToLocation,vTempLocation,fTempLength);

				//set closest spline to temp spline if the distance to 
				//the reference location is smaller than the previous
				if(closestSpline==none || fTempDist<fClosestDist)
				{
					closestSpline = checkSpline;
					fClosestDist = fTempDist;
					fClosestLength = fTempLength;
				}
			}
		}
	}

	//set the out parameters in case a spline actor was found
	if(closestSpline != none)
	{
		if(_fAcceptRadius<=0 || fClosestDist<_fAcceptRadius)
		{
			_outSpline = closestSpline;
			_outfLength = fClosestLength;
			return true;
		}
	}

	return false;
}

/**Check whether the pawn is inside a AdvKitVolume and try to go to the coresponding state
 * 
 * @return  true if a state change has occured
 */
final function bool CheckAdvVolume()
{
	local AdvKitVolume advkitVolume;

	local AdvKitSplineActor tmpSpline;
	local float fNewDistance;

	//the pawn might already be in a AdvKit volume but no spline was found while 
	//falling when NotifyPhysicsVolumeChange was called
	advkitVolume = AdvKitvolume(Pawn.PhysicsVolume);
	if(advkitVolume != none)
	{
		if(advkitVolume.bBalanceVolume)
		{
			if(GetClosestSpline(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(), class'AdvKitSplineActorBalance', tmpSpline,fNewDistance,advPawn.fMaxBalanceCheckDistance,advPawn.fMaxBalanceAcceptDistance))
			{
				GotoState('PlayerBalancing');
				return true;
			}
		}
		else if(advkitVolume.bTightspaceVolume)
		{
			if(GetClosestSpline(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(), class'AdvKitSplineActorTightSpace', tmpSpline,fNewDistance,advPawn.fMaxTightspaceCheckDistance,advPawn.fMaxTightspaceAcceptDistance))
			{
				GotoState('PlayerTightSpaceWalking');
				return true;
			}
		}
	}
	return false;
}

/**Do a trace to find out whether there is a wall beneath the ledge
 * 
 * @return  true if there is a wall, false otherwise
 */
final function bool CheckForWall()
{
	local Vector vTraceStart, vTraceEnd;

	vTraceStart = Pawn.Location;
	vTraceEnd = vTraceStart + ((vect(1,0,0)*advPawn.fWallCheckTraceLength) >> Pawn.Rotation);

	return !FastTrace(vTraceEnd, vTraceStart);
}

/**This function extends the default Trace function:
 * most parameters are optional and it has additional out parameters
 * 
 * @param   _vStartTrace        position to start the trace
 * @param   _vEndTrace          position to end the trace
 * @param   _vExtent            thickness of the trace line
 * @param   _bTraceActors       whether to trace actors or brushes
 * @param   _vOutHitLocation    location of trace hit
 * @param   _vOutHitNormal      normal of polygon the trace hit
 * @param   _outHitInfo         additional hit info
 * @param   _iExtraTraceFlags   additional trace flags (see Actor.uc TRACEFLAG_*)
 * @param   _physProperty       AdvKitPhysicalMaterialProperty of the hit material
 * @return                      first hit actor, or none if nothing was hit
 */
final function Actor TraceAdvanced(Vector _vStartTrace, Vector _vEndTrace, optional Vector _vExtent, optional bool _bTraceActors, optional out Vector _vOutHitLocation, optional out Vector _vOutHitNormal, optional out TraceHitInfo _outHitInfo, optional int _iExtraTraceFlags, optional out AdvKitPhysicalMaterialProperty _physProperty)
{
	local Actor hitActor;

	//do a standard trace
	hitActor = Trace(_vOutHitLocation, _vOutHitNormal, _vEndTrace, _vStartTrace, _bTraceActors, _vExtent, _outHitInfo, _iExtraTraceFlags);
	if(hitActor != none)
	{
		//get the material property in case there is one
		if(_outHitInfo.PhysMaterial!=none)
		{
			//make sore it is a AdvKitPhysicalMaterialProperty
			_physProperty = AdvKitPhysicalMaterialProperty(_outHitInfo.PhysMaterial.PhysicalMaterialProperty);
		}
	}

	return hitActor;
}

/**Check if current splineDistance+offset is still on spline curve. It does not check 
 * for the current spline actor only, but looks ahead to the next connection. If the
 * offset goes over the current SplineActor, this function calls NotifyPreSplineEnd
 * 
 * @param   _fCheckDistance     the offset
 * @return                      true if it's still on the spline curve
 */
final function bool CheckSplineEnd(optional float _fCheckDistance)
{
	local float fDistance;
	
	//check the connection to the next spline actor
	if(_fCheckDistance>0)
	{
		//how much distance is left
		fDistance = Abs(advPawn.fCurrentSplineDistance-advPawn.currentSpline.GetSplineLength());
		if(fDistance<_fCheckDistance)
		{
			//check if the spline has a next segment
			if(advPawn.currentSpline.HasNext() && advPawn.currentSpline.GetNext().HasNext())
			{
				NotifyPreSplineEnd(advPawn.currentSpline.GetNext(),fDistance);
			}
			else
			{
				NotifyPreSplineEnd(none,fDistance);
				//has no next segment
				return true;
			}
		}
	}

	//check the connection to the previous spline actor
	if(_fCheckDistance<0)
	{
		if(advPawn.fCurrentSplineDistance<-_fCheckDistance)
		{
			//check if the spline has a previous segment
			if(advPawn.currentSpline.HasPrevious())
			{
				NotifyPreSplineEnd(advPawn.currentSpline.GetPrevious(),advPawn.fCurrentSplineDistance);
			}
			else
			{
				NotifyPreSplineEnd(none,advPawn.fCurrentSplineDistance);
				//has no previous segment
				return true;
			}
		}
	}

	//offset is not over spline end
	return false;
}

//NOTE: not technically events! However, let's pretend they are, it makes them happy
/**Called when the Pawn's spline just changed
 * 
 * @param _oldSplineActor    the old spline
 * @param _newSplineActor    the new spline
 * @return                   true to prevent OnSplineChanged notification on the pawn
 */
event bool NotifySplineChanged(AdvKitSplineActor _oldSplineActor, AdvKitSplineActor _newSplineActor);

/**Called when the Pawn is about to change splines from CheckSplineEnd
 * 
 * @param _nextSplineActor   the next spline
 * @param _fDistToChange     how long until the change
 */
event NotifyPreSplineEnd(optional AdvKitSplineActor _nextSplineActor, optional float _fDistToChange);

/**Try to find a ledge to grab while falling down
 * 
 * @param   _outLedge       the found ledge
 * @param   _outfDistance   the distance on the spline for the found location
 * @return                  true if a ledge was found
 */
final function bool CheckLedgeGrab(out AdvKitSplineActor _outLedge, out float _outfDistance)
{
	local Vector vTraceStart, vTraceEnd;
	local Vector vHitLocation, vHitNormal;
	local Vector vForwardAxis;

	if(Pawn.Physics == PHYS_Falling && Pawn.Velocity.Z <0)
	{
		vForwardAxis = vect(1,0,0) >> Pawn.Rotation;

		vTraceStart = Pawn.Location + vForwardAxis*advPawn.fLedgeGrabTraceXOffset + vect(0,0,1)*advPawn.fLedgeGrabTraceHeight;
		vTraceEnd = vTraceStart - vect(0,0,1)*advPawn.fLedgeGrabTraceLength;

		//hit something?
		if(TraceAdvanced(vTraceStart,vTraceEnd,,,vHitLocation,vHitNormal)!=none)
		{
			//is it a 'planar' surface?
			if((vHitNormal dot vect(0,0,1)) > advPawn.fLedgeClimb_MinNormal)
			{
				//try to find a ledge next to the hit location
				if(GetClosestSpline(vHitLocation, class'AdvKitSplineActorLedge', _outLedge, _outfDistance,advPawn.fMaxLedgeCheckDistance,advPawn.fMaxLedgeAcceptDistance))
				{
					//found a ledge
					return true;
				}
			}
		}
	}

	//no ledge
	return false;
}

/**Look if the Pawn is walking over an edge and try to find a ledge to grab
 * 
 * @param   _outLedge       the found ledge
 * @param   _outfDistance   the distance on the spline for the found location
 * @return                  true if a ledge was found
 */
final function bool CheckWalkOverLedge(out AdvKitSplineActor _outLedge, out float _outfDistance)
{
	local Vector    vTraceStart, vTraceEnd;
	local Vector    X,Y,Z;

	GetAxes(Pawn.Rotation,X,Y,Z);

	vTraceStart = Pawn.Location+X*Pawn.GetCollisionRadius();
	vTraceStart.Z -= Pawn.GetCollisionHeight();
	vTraceEnd = vTraceStart + vect(0,0,-1)*16;

	//test if there is a floor
	if(FastTrace(vTraceEnd,vTraceStart))
	{
		//there is none, so try to find a ledge
		if(GetClosestSpline(vTraceStart, class'AdvKitSplineActorLedge', _outLedge, _outfDistance,advPawn.fMaxLedgeCheckDistance,advPawn.fMaxLedgeAcceptDistance))
		{
			//found a ledge
			return true;
		}
	}

	//found no ledge / Pawn is still on solid ground
	return false;
}






/*----------------------------------------------------------------------------------------*\
|*----------------------------------------Climbing----------------------------------------*|
\*----------------------------------------------------------------------------------------*/

/**Make a trace and to check if it hit a surface that is climbable
 * 
 * @param   _vTraceStart    the position to start the trace from
 * @param   _vTraceEnd      the end position of the trace
 * @param   _vOutHitNormal  the normal of the surface
 * @return                  true if a climbable surface was hit
 */
final function bool TraceForWallClimbSurface(Vector _vTraceStart, Vector _vTraceEnd, optional out Vector _vOutHitNormal)
{
	local AdvKitPhysicalMaterialProperty matProperty;

	//do the trace
	TraceAdvanced(_vTraceStart,_vTraceEnd,,,_vOutHitNormal,,,,matProperty);
	if(matProperty!=none)
	{
		//only check the property
		if(matProperty.bIsClimbableWall)
		{
			return true;
		}
	}

	//not hit a climbable surface
	return false;
}


/**Check if the Pawn can climb into the given direction
 * 
 * @param   _vOutDirection      the direction the player want to climb in, will be modified by the function
 *                              when partial movement is possible
 * @return                      true if the it is possible to climb in the direction
 */
final function bool CheckForClimbLocationInDirection(out Vector _vOutDirection)
{
	local float fTraceDepth;
	local Vector vXOffset,vYOffset,vZOffset;
	local Vector X,Y,Z;
	local Vector vTraceStart1, vTraceEnd1, vTraceStart2, vTraceEnd2;
	local Vector vGlobalWallNormal;
	local Vector vClimbUpNoWallTraceStart, vClimbUpNoWallTraceEnd;
	local Vector vClimbUpTopPositionTraceStart, vClimbUpTopPositionTraceEnd;
	local Vector vHitLocation, vHitNormal, vScaleUpStartLocation;

	fTraceDepth = advPawn.fWallTraceLength;
	GetAxes(Pawn.Rotation,X,Y,Z);

	vXOffset = X*Pawn.GetCollisionRadius();
	vYOffset = Y*Pawn.GetCollisionRadius();
	vZOffset = Z*Pawn.GetCollisionHeight();


	//Player is climbing up
	if(_vOutDirection.Z > 0)
	{
		//Check for climbable ceiling
		if(TraceForWallClimbSurface(Pawn.Location,Pawn.Location+Z*advPawn.fCeilingTraceLength))
		{
			advPawn.wallClimbingState = EWCS_ToCeiling;
			advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
			return false;
		}

		vTraceStart1 = Pawn.Location+vZOffset + vYOffset + Z*_vOutDirection.Z;
		vTraceEnd1 = vTraceStart1+X*fTraceDepth;
			
		vTraceStart2 = Pawn.Location+vZOffset - vYOffset + Z*_vOutDirection.Z;
		vTraceEnd2 = vTraceStart2+X*fTraceDepth;

		//Not climbable in this direction?
		if(!TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) || !TraceForWallClimbSurface(vTraceStart2,vTraceEnd2))
		{
			_vOutDirection.Z = 0;

			vClimbUpNoWallTraceStart = Pawn.Location + vect(0,0,1)*advPawn.fWallClimbClimbUpTraceHeight;
			vClimbUpNoWallTraceEnd = vClimbUpNoWallTraceStart + ((vect(1,0,0)*advPawn.fWallClimbClimbUpTraceLength) >> Pawn.Rotation);

			//is there no obstacle in the up direction?
			if(FastTrace(vClimbUpNoWallTraceStart,vClimbUpNoWallTraceEnd))
			{
				//Check whehter the pawn can stand on top
				vClimbUpTopPositionTraceStart = Pawn.Location + vXOffset*2 + vZOffset*2;
				vClimbUpTopPositionTraceEnd = vClimbUpTopPositionTraceStart - vZOffset*2;

				if(Trace(vHitLocation, vHitNormal,vClimbUpTopPositionTraceEnd, vClimbUpTopPositionTraceStart,true)!=none)
				{
					//can the pawn stand on top?
					if(advPawn.fWallClimbTopMinNormal <= (vHitNormal dot vect(0,0,1)))
					{
						//set the pawn to be exactly on the edge for the roo animation to fit
						vScaleUpStartLocation = Pawn.Location;
						vScaleUpStartLocation.Z = vHitLocation.Z - Pawn.GetCollisionHeight() ;
						Pawn.SetRelativeLocation(advPawn.Global2RelativePosition(vScaleUpStartLocation));

						advPawn.bCollideWorld = false;
						advPawn.wallClimbingState = EWCS_ScaleUp;
						advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);

						return false;
					}
				}
			}
		}
	}

	//Player is climbing down
	if(_vOutDirection.Z < 0)
	{
		vTraceStart1 = Pawn.Location-vZOffset + vYOffset + Z*_vOutDirection.Z;
		vTraceEnd1 = vTraceStart1+X*fTraceDepth;
			
		vTraceStart2 = Pawn.Location-vZOffset - vYOffset + Z*_vOutDirection.Z;
		vTraceEnd2 = vTraceStart2+X*fTraceDepth;

		if(!TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) || !TraceForWallClimbSurface(vTraceStart2,vTraceEnd2))
		{
			_vOutDirection.Z = 0;
		}
	}

	//Player is climbing right
	if(_vOutDirection.Y > 0)
	{
		vTraceStart1 = Pawn.Location+vZOffset + vYOffset + Y*_vOutDirection.Y;
		vTraceEnd1 = vTraceStart1+X*fTraceDepth;
			
		vTraceStart2 = Pawn.Location-vZOffset + vYOffset + Y*_vOutDirection.Y;
		vTraceEnd2 = vTraceStart2+X*fTraceDepth;

		//Not climbable in this direction?
		if(!TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) || !TraceForWallClimbSurface(vTraceStart2,vTraceEnd2))
		{
			_vOutDirection.Y = 0;

			vTraceStart1 = Pawn.Location+advPawn.v2WallEdgeTraceNear.Y*Y+advPawn.v2WallEdgeTraceNear.X*X;
			vTraceEnd1 = Pawn.Location+advPawn.v2WallEdgeTraceNear.X*X;

			vTraceStart2 = Pawn.Location+advPawn.v2WallEdgeTraceFar.Y*Y+advPawn.v2WallEdgeTraceFar.X*X;
			vTraceEnd2 = Pawn.Location+advPawn.v2WallEdgeTraceFar.X*X;

			//Check for wall edge
			vGlobalWallNormal = advPawn.vCurrentWallNormal >> Pawn.Base.Rotation;
			if(TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) && TraceForWallClimbSurface(vTraceStart2,vTraceEnd2, vGlobalWallNormal))
			{
				advPawn.vCurrentWallNormal = vGlobalWallNormal << Pawn.Base.Rotation;
				advPawn.bCollideWorld = false;
				advPawn.wallClimbingState = EWCS_AroundCornerRight;
				advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
			}
		}
	}

	//Player is climbing left
	if(_vOutDirection.Y < 0)
	{
		vTraceStart1 = Pawn.Location+vZOffset - vYOffset + Y*_vOutDirection.Y;
		vTraceEnd1 = vTraceStart1+X*fTraceDepth;
			
		vTraceStart2 = Pawn.Location-vZOffset - vYOffset + Y*_vOutDirection.Y;
		vTraceEnd2 = vTraceStart2+X*fTraceDepth;

		if(!TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) || !TraceForWallClimbSurface(vTraceStart2,vTraceEnd2))
		{
			_vOutDirection.Y = 0;

			vTraceStart1 = Pawn.Location-advPawn.v2WallEdgeTraceNear.Y*Y+advPawn.v2WallEdgeTraceNear.X*X;
			vTraceEnd1 = Pawn.Location+advPawn.v2WallEdgeTraceNear.X*X;

			vTraceStart2 = Pawn.Location-advPawn.v2WallEdgeTraceFar.Y*Y+advPawn.v2WallEdgeTraceFar.X*X;
			vTraceEnd2 = Pawn.Location+advPawn.v2WallEdgeTraceFar.X*X;

			//Check for wall edge
			vGlobalWallNormal = advPawn.vCurrentWallNormal >> Pawn.Base.Rotation;
			if(TraceForWallClimbSurface(vTraceStart1,vTraceEnd1) && TraceForWallClimbSurface(vTraceStart2,vTraceEnd2, vGlobalWallNormal))
			{
				advPawn.vCurrentWallNormal = vGlobalWallNormal << Pawn.Base.Rotation;
				advPawn.bCollideWorld = false;
				advPawn.wallClimbingState = EWCS_AroundCornerLeft;
				advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
			}
		}
	}

	return true;
}

/**Make a trace to check if the direction the player wants to climb 
 * and check whether it's possible
 * 
 * @param   _vDirection     the direction to climb in
 * @return                  true if the Pawn can climb in this direction
 */
final function bool CheckForCeilingLocationInDirection(Vector _vDirection)
{
	local Vector vXOffset,vYOffset,vZOffset;
	local Vector vTraceStartRight, vTraceEndRight;
	local Vector vTraceStartLeft, vTraceEndLeft;

	//if the pawn is not moving, no trace is necessary
	if(VSize(_vDirection)==0)
	{
		return false;
	}

	GetAxes(Rotator(_vDirection),vXOffset,vYOffset,vZOffset);

	vXOffset = vXOffset*Pawn.GetCollisionRadius();
	vYOffset = vYOffset*Pawn.GetCollisionRadius();
	vZOffset = vZOffset*Pawn.GetCollisionHeight();

	//trace on left and right side of Pawn
	vTraceStartRight = Pawn.Location+vZOffset + vYOffset + vXOffset;
	vTraceEndRight = vTraceStartRight + vZOffset;
			
	vTraceStartLeft = Pawn.Location+vZOffset - vYOffset + vXOffset;
	vTraceEndLeft = vTraceStartLeft + vZOffset;

	//do traces
	if(!TraceForWallClimbSurface(vTraceStartRight,vTraceEndRight) || !TraceForWallClimbSurface(vTraceStartLeft,vTraceEndLeft))
	{
		//no climbable surface could be found
		return false;
	}

	//can climb in direction
	return true;
}

/**Check for a floor beneath the pawn
 * @return  true if there is a floor
 */
final function bool CheckForFloor()
{
	local Vector vTraceStart, vTraceEnd;

	vTraceStart = Pawn.Location;
	vTraceEnd = vTraceStart - vect(0,0,1)*advPawn.fFloorTraceLength;
	return !FastTrace(vTraceEnd,vTraceStart);
}

/**This function is called from a timer inside of the PlayerWallJumping state
 * to make the Pawn fall down after a time specified by fWallJumpFallDownTime
 */
final function FallDown() 
{
	GotoState('PlayerWalking');
}


/**Called when our pawn has collided with a blocking piece of world geometry
 *
 *@param _vHitNormal    the normal of the hit surface
 *@param _wall          the hit wall
 *@return               false to call HitWall on the pawn, true to prevent it
 */
event bool NotifyHitWall(vector _vHitNormal, actor _wall)
{
	local Vector    vTraceStart, vTraceEnd;
	local Vector    vHitLocation, vTraceHitNormal;
	local Actor     hitActor;
	local AdvKitPhysicalMaterialProperty physMatProperty;
	local Vector    X,Y,Z;
	local Vector    vWallJumpTopTraceStart, vWallJumpTopTraceEnd, vWallJumpBottomTraceStart, vWallJumpBottomTraceEnd;

	//cannot work if there is no advPawn
	if(advPawn == none)
	{
		return false;
	}

	//hit a pole
	if(_wall.IsA('AdvKitPole'))
	{
		//set the current pole
		advPawn.SetCurrentPole(AdvKitPole(_wall));
		if(Pawn.Location.Z+Pawn.GetCollisionHeight() > advPawn.currentPole.Location.Z + advPawn.currentPole.GetCollisionHeight())
		{
			//the player is so high that he can stand on the pole
			GotoState('PlayerPoleStanding');
		}
		else
		{
			//he is not high enough to stand on the pole
			GotoState('PlayerPoleClimbing');
		}
		return false;
	}

	//do a trace to find out wheter we can climb on the hit object
	vTraceStart = Pawn.Location;
	vTraceEnd = vTraceStart - Normal(_vHitNormal)*advPawn.fWallTraceLength;

	hitActor = TraceAdvanced(vTraceStart, vTraceEnd,,,vHitLocation, vTraceHitNormal,,,physMatProperty);
	if(hitActor!=none && physMatProperty!=none)
	{
		if(physMatProperty.bIsClimbableWall)
		{
			//climb on the wall
			advPawn.SetCurrentWall(hitActor, vTraceHitNormal);
			GotoState('PlayerWallClimbing');
			return false;
		}
		else if(physMatProperty.bIsWallJumpingPossible)
		{
			//only wall jump if the pawn is falling (and not walking)
			if(Pawn.Physics == PHYS_Falling)
			{
				GetAxes(Pawn.Rotation,X,Y,Z);
				vWallJumpTopTraceStart = Pawn.Location + Z*Pawn.GetCollisionHeight();
				vWallJumpTopTraceEnd = vWallJumpTopTraceStart + X*Pawn.GetCollisionRadius()*2;
					
				vWallJumpBottomTraceStart = Pawn.Location - Z*Pawn.GetCollisionHeight();
				vWallJumpBottomTraceEnd = vWallJumpBottomTraceStart + X*Pawn.GetCollisionRadius()*2;

				//chekc whether there is actually a wall at the top and bottom of the pawn's collision extend
				if(!FastTrace(vWallJumpTopTraceEnd, vWallJumpTopTraceStart) && !FastTrace(vWallJumpBottomTraceEnd, vWallJumpBottomTraceStart))
				{
					advPawn.SetCurrentWall(hitActor, vTraceHitNormal);
					GotoState('PlayerWallJumping');
					return false;
				}
			}
		}
	}

	//nothing happened
	return false;
}

/**Called when our pawn has landed on a floor
 *
 *@param _vHitNormal    the normal of the hit surface
 *@param _floor         the hit floor
 *@return               false to call Landed on the pawn, true to prevent it
 */
event bool NotifyLanded(vector _vHitNormal, Actor _floor)
{

	//landed on a pole
	if(advPawn!=none && _floor.IsA('AdvKitPole'))
	{
		advPawn.SetCurrentPole(AdvKitPole(_floor));
		GotoState('PlayerPoleStanding');
	}

	CheckAdvVolume();

	return super.NotifyLanded(_vHitNormal, _floor);
}

/**Called to recalculate the Pawn's location on a climbable surface
 */
final function RecalculateClimbingPosition()
{
	local Vector    vHitLocation, vHitNormal;
	local Vector    vNewLocation;
	local Actor     newBase;
	local AdvKitPhysicalMaterialProperty matProp;

	//do a trace to find the next location on the surface to set the pawn to
	newBase = TraceAdvanced(Pawn.Location,Pawn.Location+(vect(1,0,0) >> Pawn.Rotation)*Pawn.GetCollisionRadius()*2,,,vHitLocation,vHitNormal,,,matProp);
	//can only climb on a climbable surface
	if(matProp!=none && matProp.bIsClimbableWall)
	{
		vNewLocation = vHitLocation+vHitNormal*Pawn.GetCollisionRadius();
		advPawn.SetCurrentWall(newBase,vHitNormal);
		Pawn.SetRelativeLocation(advPawn.Global2RelativePosition(vNewLocation));
	}
}











/*----------------------------------------------------------------------------------------*\
|*-----------------------------------------Misc Functions---------------------------------*|
\*----------------------------------------------------------------------------------------*/

/**The controller posesses a Pawn
 * 
 * @param   _inPawn                 the Pawn that is getting possesed
 * @param   _bVehicleTransition     is it a vehicle transition (not that important)
 */
event Possess(Pawn _inPawn, bool _bVehicleTransition)
{
	//cast the pawn into AdvKitPawn
	advPawn = AdvKitPawn(_inPawn);

	super.Possess(_inPawn, _bVehicleTransition);
}

/**Default behaviour (for walking) to rotate the pawn. This is called from 
 * UpdateRotation when the advPawn is not in RootMotion.
 * 
 * @param   _fDeltaTime     the delta time
 */
function UpdatePawnRotation(float _fDeltaTime)
{
	local Vector vDirection;

	//rotate the pawn only if it is not in the air
	if(Pawn.Physics != PHYS_Falling)
	{
		//only rotate around the Z axis
		vDirection = Pawn.Acceleration;
		vDirection.Z = 0;

		//only rotate if the Pawn is actually being accelerated, 
		//otherwise Rotator(vDirection) will be rot(0,0,0)
		if(VSize(vDirection)>0)
		{
			Pawn.SetRotation(RLerp(Pawn.Rotation,Rotator(vDirection),FMin(_fDeltaTime*advPawn.fRotationLerpSpeed,1),true));
		}
	}
}

/**Update all rotation for Controller, Camera and Pawn
 * 
 * @param   _fDeltaTime     the delta time
 */
function UpdateRotation( float _fDeltaTime )
{
	local Rotator	rDeltaRot, rViewRotation;

	//Manage and rotate the camera
	rViewRotation = Rotation;

	// Calculate Delta to be applied on ViewRotation
	rDeltaRot.Yaw	= PlayerInput.aTurn;
	rDeltaRot.Pitch	= PlayerInput.aLookUp;

	ProcessViewRotation( _fDeltaTime, rViewRotation, rDeltaRot );
	SetRotation(rViewRotation);

	ViewShake( _fDeltaTime );


	if(advPawn==none)
		return;

	//only rotate the pawn when it is not in RootMotion
	if(!advPawn.bInRootRotation)
	{
		UpdatePawnRotation(_fDeltaTime);
	}
}

/**Zoom the camera (modify distance of the camera to the Pawn)
 * 
 * @param   _fZoom          the distance amount to zoom
 * @param   _fDeltaTime     the delta time
 */
final exec function Zoom(float _fZoom, optional float _fDeltaTime)
{
	local float fNewZoom;
	local float fScale;

	//set scale to 1 in case delta time is zero
	fScale = _fDeltaTime == 0 ? 1.0f : _fDeltaTime;

	//calculate the zoom
	fNewZoom = PlayerCamera.FreeCamDistance += fScale * _fZoom;
	fNewZoom = FClamp(fNewZoom,fMinCameraDistance,fMaxCameraDistance);

	//set the zoom
	PlayerCamera.FreeCamDistance = fNewZoom;
}

/**Called from Pawn.OnAnimEnd when the pawn finished playing an animation
 * 
 * Look at Actor.OnAnimEnd for Documentation
 */
event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime);

/**
 * Called from AdvKitLeverTrigger upon use to have the advPawn play a synchronized 
 * use animation
 * 
 * @param   _interaction    the state of the interaction (animation controll of the AnimTree)
 */
final function Interact(EInteractionState _interaction)
{
	advPawn.interactionState = EIS_None;
	advPawn.interactionState = _interaction;
	GotoState('PlayerInteracting');
}






/*----------------------------------------------------------------------------------------*\
|*----------------------------------------Player States-----------------------------------*|
\*----------------------------------------------------------------------------------------*/

state PlayerWalking
{
	event BeginState(name PreviousStateName)
	{
		super.BeginState(PreviousStateName);
		
		Pawn.SetHardAttach(false);
		Pawn.SetCollision(true,true,false);

		if(advPawn==none)
		{
			return;
		}

		advPawn.adventureState = EAS_Default;
	}

	event EndState(name NextStateName)
	{
		super.EndState(NextStateName);

		//make sure to undo any modification
		Pawn.MovementSpeedModifier = 1;
	}

	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		if(advPawn.bVaulting)
		{
			//return to normal movement after vaulting
			advPawn.bVaulting = false;
			Pawn.SetPhysics(PHYS_Walking);
			Pawn.bCollideWorld = true;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector    vNormalInput;
		local Vector	vNewAccel;
		local float     fInputStrength;
		local AdvKitSplineActor tmpLedge;
		local float fNewDistance;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			// Update rotation.
			UpdateRotation( _fDeltaTime );

			//only update if the pawn is not in root motion
			if(!advPawn.bInRootMotion)
			{

				vNormalInput = GetRelativeInputMovement(_fDeltaTime);
				vNormalInput.Z = 0;
				fInputStrength = FClamp(VSize(vNormalInput),0,1);

				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime) * Pawn.AccelRate;

				//each of these two lines guarantee analog movement speed, I do not really like both 
				//solutions but I did not find any others
				//Pawn.GroundSpeed = Pawn.default.GroundSpeed*fInputStrength;
				Pawn.MovementSpeedModifier = fInputStrength;
				Pawn.Acceleration = vNewAccel;				
				
				if(bPressedJump)
				{
					if(Pawn.Physics == PHYS_Walking)
					{
						Pawn.DoJump(bUpdating);
					}
					bPressedJump = false;
				}

				if(Pawn.Physics == PHYS_Walking)
				{
					//try to vault
					if(VSize(Pawn.Acceleration)>0 && CheckVaulting())
					{
						advPawn.bCollideWorld = false;
						advPawn.SetPhysics(PHYS_None);
						advPawn.ActivateRootMotion();
						advPawn.bVaulting = true;
					}

					//grab onto ledge if possible when walking over an edge
					if(CheckWalkOverLedge(tmpLedge, fNewDistance))
					{
						advPawn.SetCurrentSpline(tmpLedge,fNewDistance);
						advPawn.ledgeClimbingState = ELCS_FallingLedgeGrab;
						advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);

						Pawn.SetPhysics(PHYS_Interpolating);
						Pawn.SetRelativeRotation(Rotator(advPawn.vCurrentSplineRelativeNormal));

						GotoState('PlayerLedgeClimbing');
						return;
					}

					//start sliding if possible
					if(CheckIfOnSlide())
					{
						GotoState('PlayerSliding');
					}

					if(CheckAdvVolume())
						return;
				}

				//check if the pawn can grab any ledges
				if(CheckLedgeGrab(tmpLedge, fNewDistance))
				{
					advPawn.SetCurrentSpline(tmpLedge,fNewDistance);
					GotoState('PlayerLedgeClimbing');
				}
			}

			//ignore jumps in case the pawn is in root motion
			bPressedJump = false;
		}
	}
}

state PlayerPoleClimbing
{
	event BeginState(name PreviousStateName)
	{
		if(advPawn!=none)
		{
			advPawn.poleClimbingState  = EWPS_ClimbPole;
			//Since the location will be set manually we disable collision 
			Pawn.SetPhysics(PHYS_Interpolating);
			advPawn.adventureState = EAS_PoleClimbing;
		}
	}

	event EndState(name NextStateName)
	{
		if(NextStateName!='PlayerPoleStanding')
		{
			advPawn.SetBase(none);
			advPawn.SetHardAttach(false);
			Pawn.SetPhysics(PHYS_Falling);
		}
	}


	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.poleClimbingState)
		{
		case EWPS_ClimpUp:
			GotoState('PlayerPoleStanding');
			break;
		case EWPS_JumpBack:
			GotoState('PlayerWalking');
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector    X, Y, Z;
		local Vector    vDeltaLoc, vNewAccel, vDir;
		local Rotator   rDeltaRot;

		if(advPawn==none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion && !advPawn.bInRootRotation)
			{
				vNewAccel = GetRelativeInputMovement(_fDeltaTime);

				//Climb upwards
				if(vNewAccel.X>0)
				{
					//check for a ceiling
					if(FastTrace(Pawn.Location+vect(0,0,1)*advPawn.fCeilingTraceLength,Pawn.Location))
					{
						if(Pawn.RelativeLocation.Z + Pawn.GetCollisionHeight() >= advPawn.currentPole.GetCollisionHeight())
						{
							//climb up on top of the pole
							advPawn.poleClimbingState  = EWPS_ClimpUp;
							advPawn.ActivateRootMotion(RMM_Relative);
							return;
						}
					}
					else
					{
						vNewAccel.X=0;
					}
				}
				//climb down
				else if(vNewAccel.X<0)
				{
					//pawn climbed lower than the pole is -> fall down
					if(Pawn.RelativeLocation.Z - Pawn.GetCollisionHeight() <= -advPawn.currentPole.GetCollisionHeight())
					{
						GotoState('PlayerWalking');
					}
					//pawn hit the floor, go back to walking
					else if(CheckForFloor())
					{
						GotoState('PlayerWalking');
					}
				}

				//blend pawn animations depending on direction
				vDir = Normal(vNewAccel);
				advPawn.fAnimTreeFrontDirectionBlend = Lerp(advPawn.fAnimTreeFrontDirectionBlend, vect(1,0,0) dot vDir,_fDeltaTime*10);
				advPawn.fAnimTreeSideDirectionBlend = Lerp(advPawn.fAnimTreeSideDirectionBlend, vect(0,1,0) dot vDir,_fDeltaTime*10);

				vDeltaLoc = Pawn.RelativeLocation;
				vDeltaLoc.Z = 0;

				rDeltaRot = Rotator(vDeltaLoc);
				rDeltaRot.Yaw -= vNewAccel.Y*_fDeltaTime*DegToUnrRot*advPawn.fClimbDegreePerSec;
				GetAxes(rDeltaRot, X, Y, Z);


				//NOTE: Use this line to prevent the pawn from "skipping" when starting to climb (does not work if the pole is already a based actor)
				//Pawn.SetRelativeLocation(advPawn.Global2RelativePosition(Pawn.Location)); 

				vDeltaLoc = X*(advPawn.fClimbPoleDistance+advPawn.currentPole.GetCollisionRadius());
				vDeltaLoc.Z = Pawn.RelativeLocation.Z;
				vDeltaLoc.Z += vNewAccel.X*advPawn.fClimbPoleSpeed*_fDeltaTime;
				Pawn.SetRelativeLocation(vDeltaLoc);


				//Jump, obviously
				if (bPressedJump && advPawn.poleClimbingState==EWPS_PreBackJump)
				{
					Pawn.DoJump(bUpdating);
					bPressedJump = false;
					return;
				}

				//the player pressed crouch, prepare for jumping back
				if(PlayerInput.aUp < 0)
				{
					advPawn.poleClimbingState = EWPS_PreBackJump;
				}
				else
				{
					//reset
					if(advPawn.poleClimbingState != EWPS_ClimbPole)
					{
						advPawn.poleClimbingState = EWPS_ClimbPole;
					}
				}
			}

			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		local Vector    vDeltaLoc;

		//make the pawn face the pole
		vDeltaLoc = -Pawn.RelativeLocation;
		vDeltaLoc.Z = 0;

		Pawn.SetRelativeRotation(Rotator(vDeltaLoc));
	}
}

state PlayerPoleStanding
{
	event BeginState(name PreviousStateName)
	{
		if(advPawn!=none)
		{
			advPawn.poleClimbingState  = EWPS_StandOnPole;

			Pawn.SetPhysics(PHYS_Interpolating);
			advPawn.adventureState = EAS_PoleClimbing;

			advPawn.SetRelativeLocation(vect(0,0,1)*(advPawn.currentPole.GetCollisionHeight()+advPawn.fStandPoleHeight));
		}
	}

	event EndState(name NextStateName)
	{
		//do not hold the pole e.g. in case the player jumps off
		if(NextStateName!='PlayerPoleClimbing')
		{
			if(advPawn!=none)
			{
				advPawn.SetCurrentPole(none);
			}
		}
	}

	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.poleClimbingState)
		{
		case EWPS_ClimbDown:
			GotoState('PlayerPoleClimbing');
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion)
			{
				Pawn.ZeroMovementVariables();

				//Jump, obviously
				if (bPressedJump)
				{
					Pawn.DoJump(bUpdating);
					bPressedJump = false;
					GotoState('PlayerWalking');
				}

				//climb down the pole
				if(PlayerInput.aUp < 0)
				{
					advPawn.poleClimbingState  = EWPS_ClimbDown;
					advPawn.ActivateRootMotion(RMM_Relative);
				}
			}

			// Update rotation.
			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdatePawnRotation( float _fDeltaTime )
	{
		local Vector	vNewAccel;

		//face the in the direction the player is steering to
		vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);
		if(VSize(vNewAccel)!=0)
		{
			vNewAccel.Z = 0;
			vNewAccel = advPawn.Global2RelativeDirection(vNewAccel);
			Pawn.SetRelativeRotation(RLerp(Pawn.RelativeRotation,Rotator(vNewAccel),FMin(1,advPawn.fRotationLerpSpeed*_fDeltaTime),true));
		}
	}
}

state PlayerSwimming
{
	event BeginState(name PreviousStateName)
	{
		super.BeginState(PreviousStateName);

		//set the camera mode
		if(PlayerCamera.IsA('AdvKitCamera'))
		{
			Camera('Swimming');
		}

		advPawn.swimmingState = EWS_Swimming;
		advPawn.adventureState = EAS_Swimming;
	}

	event EndState(name NextStateName)
	{
		//reset collision cylinder, because it might have been modified
		Pawn.CylinderComponent.SetCylinderSize(advPawn.GetCollisionRadiusDefault(),advPawn.GetCollisionHeightDefault());

		//reset the camera mode
		if(PlayerCamera.IsA('AdvKitCamera'))
		{
			Camera('ThirdPerson');
		}

	}


	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.swimmingState)
		{
		case EWS_StartDiving:
			//Start blocking the camera again
			if(PlayerCamera.IsA('AdvKitCamera'))
			{
				AdvKitCamera(PlayerCamera).bCollideWithWaterSurface=true;
			}
			GotoState('PlayerDiving');
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector      vSurfaceLoc;
		local Vector vNewAccel;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion)
			{
				//Cannot jump while swimming
				bPressedJump = false;

				//player pushed any down button/axis, start diving
				if(PlayerInput.aUp<0)
				{
					//Make the camera go through the surface while starting to dive
					if(PlayerCamera.IsA('AdvKitCamera'))
					{
						 AdvKitCamera(PlayerCamera).bCollideWithWaterSurface=false;
					}
					Pawn.ZeroMovementVariables();
					advPawn.swimmingState = EWS_StartDiving;
					advPawn.ActivateRootMotion(RMM_Translate);
					return;
				}

				//make pawn float on surface
				vSurfaceLoc = Pawn.Location;
				vSurfaceLoc.Z = Lerp(Pawn.Location.Z,advPawn.fWaterSurfaceZ,FMin(5*_fDeltaTime,1));
				Pawn.SetLocation(vSurfaceLoc);

				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);
				vNewAccel.Z = 0;
				Pawn.Acceleration = vNewAccel*Pawn.AccelRate;

				if(VSize(Pawn.Acceleration) > 0)
				{
					//swap cylinder height and radius if pawn is swimming forward
					Pawn.CylinderComponent.SetCylinderSize(advPawn.GetCollisionHeightDefault(),advPawn.GetCollisionRadiusDefault());
				}
				else
				{
					//reset sizes when pawn stops swimming
					Pawn.CylinderComponent.SetCylinderSize(advPawn.GetCollisionRadiusDefault(),advPawn.GetCollisionHeightDefault());
				}

				UpdateRotation( _fDeltaTime );
			}
		}
	}
}

state PlayerDiving extends PlayerSwimming
{
	event BeginState(Name PreviousStateName)
	{
		if(advPawn==none)
		{
			return;
		}
		advPawn.swimmingState = EWS_Diving;

		//set the camera mode
		if(PlayerCamera.IsA('AdvKitCamera'))
		{
			Camera('Diving');
		}

		advPawn.adventureState = EAS_Swimming;
	}

	event EndState(Name NextStateName)
	{
		//reset collision cylinder, because it might have been modified
		Pawn.CylinderComponent.SetCylinderSize(advPawn.GetCollisionRadiusDefault(),advPawn.GetCollisionHeightDefault());

		//diving directly accesses the MeshComponent's rotation, reset it when leaving 
		Pawn.Mesh.SetRotation(rot(0,0,0));
		Pawn.ZeroMovementVariables();

		if(PlayerCamera.IsA('AdvKitCamera'))
		{
			Camera('ThirdPerson');
		}
	}

	function Vector GetDirectionalizedInputMovement(float _fDeltaTime)
	{
		local Vector	X,Y,Z,vInputVector;

		vInputVector = GetRelativeInputMovement(_fDeltaTime);

		//prevent speed hack (e.g. Forward is slower than Foward + Right movement)
		if(VSize(vInputVector)>1)
			vInputVector = Normal(vInputVector);

		//align acceleration to view angle
		GetAxes(PlayerCamera.Rotation,X,Y,Z);

		//final input: project X and Y to camera space and leave Z in world space
		vInputVector = Normal(vInputVector.X*X + vInputVector.Y*Y + vect(0,0,1)*vInputVector.Z);
		return vInputVector;
	}

	function PlayerMove(float _fDeltaTime)
	{
		if (advPawn == None)
		{
			GotoState('Dead');
		}
		else
		{
			//has the pawn finished the diving start animation?
			if(!advPawn.bInRootMotion)
			{
				//set acceleration directly, udk handles swimming movement
				Pawn.Acceleration = GetDirectionalizedInputMovement(_fDeltaTime) * Pawn.AccelRate;

				UpdateRotation( _fDeltaTime );

				//can't jump while diving
				bPressedJump = false;

				//check if the pawn is resurfacing
				if((Pawn.Velocity.Z>0 || Pawn.Acceleration.Z>0) && Pawn.Location.Z + Pawn.GetCollisionHeight() > advPawn.fWaterSurfaceZ)
				{
					GotoState('PlayerSwimming');
				}
			}
		}
	}
	
	function UpdatePawnRotation( float DeltaTime )
	{
		local Vector tmpDirection, vTmpMeshDirection;
		local Vector vCollisionRot;
		local float fColHeight, fColRadius, fPrevColHeight, fPrevColRadius;

		//first rotate the pawn according to the acceleration projected onto the XY plane
		tmpDirection = Pawn.Acceleration;
		tmpDirection.Z = 0;
		if(VSize(tmpDirection)!=0)
		{
			Pawn.SetRotation(RLerp(Pawn.Rotation,Rotator(tmpDirection),FMin(1,advPawn.fRotationLerpSpeed*DeltaTime),true));
		}

		//secondly rotate the pawn's mesh local pitch
		if(Pawn.Acceleration.Z!=0)
		{
			//Z is non-null so we can calculate an angle 
			vTmpMeshDirection.Z = Pawn.Acceleration.Z;
			vTmpMeshDirection.X = Sqrt(Pawn.Acceleration.X ** 2 + Pawn.Acceleration.Y ** 2);
			Pawn.Mesh.SetRotation(RLerp(Pawn.Mesh.Rotation,Rotator(vTmpMeshDirection),FMin(1,advPawn.fRotationLerpSpeed*DeltaTime),true));
		}
		else
		{
			//Z is null so we reset the mesh's rotation
			Pawn.Mesh.SetRotation(RLerp(Pawn.Mesh.Rotation,rot(0,0,0),FMin(1,advPawn.fRotationLerpSpeed*DeltaTime),true));
		}

		//Thirdly modify the collision cylinder to match the mesh's rotation and prevent clipping
		//default values
		fColHeight = advPawn.GetCollisionHeightDefault();
		fColRadius = advPawn.GetCollisionRadiusDefault();
		
		//vector representing the angle
		vCollisionRot = Normal(Pawn.Acceleration);
		//only modify default values in case the pawn is moving
		if(VSize(vCollisionRot)!=0)
		{		
			fPrevColHeight = fColHeight;
			fPrevColRadius = fColRadius;

			//the cylinder gets wider and smaller the more the player is swimming on the xy plane and
			//thinner and taller the more the player is swimming upright
			fColHeight = fPrevColHeight * Abs(vCollisionRot.Z) + fPrevColRadius * (1-Abs(vCollisionRot.Z));
			fColRadius = fPrevColRadius * Abs(vCollisionRot.Z) + fPrevColHeight * (1-Abs(vCollisionRot.Z));
		}
		
		//apply the values
		Pawn.CylinderComponent.SetCylinderSize(fColRadius, fColHeight);

	}
}

state PlayerSliding
{
	event BeginState(name PreviousStateName)
	{
		advPawn.adventureState = EAS_Sliding;
		Pawn.SetPhysics(PHYS_Flying);
		//NOTE: The collision cylinder has an offset in PlayerWalking of 3 apparently, this compensates for it
		Pawn.SetCollisionSize(Pawn.GetCollisionRadius(),advPawn.GetCollisionHeightDefault()+3);
	}

	event EndState(name NextStateName)
	{
		//reset collision size
		Pawn.SetCollisionSize(Pawn.GetCollisionRadius(),advPawn.GetCollisionHeightDefault());
	}

	function PlayerMove(float DeltaTime)
	{
		local Vector vNewAccel, vPlayerDirection;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			//Trace to find slide
			if(CheckIfOnSlide())
			{
				//Default physics on slide is PHYS_Flying, so check whether the pawn is jumping
				if(Pawn.Physics!=PHYS_Flying)
				{
					//If the pawn is not falling, set physics to flying
					if(Pawn.Physics != PHYS_Falling)
					{
						vNewAccel = vect(0,0,0); //Prevent the pawn from doing a super fast leap upon physics change
						Pawn.SetPhysics(PHYS_Flying);
					}
				}
				else
				{
					vNewAccel = GetDirectionalizedInputMovement(DeltaTime);

					//Project input onto slope and prevent 'upwards' acceleration
					vPlayerDirection = vNewAccel - FMin(vNewAccel Dot Normal(advPawn.vSlopeTangent), 0.00) * Normal(advPawn.vSlopeTangent);
					//Acceleration = slope direction + projected player input + 'gravity' vector
					vNewAccel =  advPawn.vSlopeTangent+vPlayerDirection*advPawn.fSlideControlStrength+vect(0,0,-1);

					//Slow down upward movement / accelerate downward movement
					Pawn.Velocity = Pawn.Velocity - FMin(Pawn.Velocity Dot advPawn.vSlopeNormal, 0.00) * advPawn.vSlopeNormal;

 
					//Jump, obviously
					if (bPressedJump == true)
					{
						//Cancel other input related acceleration
						vNewAccel = vect(0,0,0); 

						bPressedJump = false;
						Pawn.DoJump(bUpdating);
					}
				}
			}
			//no slide found
			else
			{
				//No slide and not falling means the pawn must have landed in or on something that is no slide
				if(Pawn.Physics != PHYS_Falling)
				{
					GotoState('PlayerWalking');
					return;
				}
			}

			Pawn.Acceleration=vNewAccel*Pawn.AccelRate;

			// Update rotation.
			UpdateRotation( DeltaTime );
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		local Vector tmpDirection;

		//Align pawn to XY-Projected Velocity
		tmpDirection = Pawn.Velocity;
		tmpDirection.Z = 0;
		if(VSize(tmpDirection)!=0)
		{
			Pawn.SetRotation(Rotator(tmpDirection));
		}
	}
}

state PlayerWallClimbing
{
	event BeginState(name PreviousStateName)
	{
		if(advPawn == none)
		{
			GotoState('PlayerWalking');
			return;
		}

		Pawn.ZeroMovementVariables();
		Pawn.SetPhysics(PHYS_Interpolating);
		advPawn.adventureState = EAS_WallClimbing;
		advPawn.wallClimbingState = EWCS_Default;
		RecalculateClimbingPosition();
	}

	event EndState(name NextStateName)
	{
		if(NextStateName!='PlayerCeilingClimbing')
		{
			advPawn.SetCurrentWall(none);
			Pawn.SetPhysics(PHYS_Falling);
		}

		//Setting PHYS_Falling when transitioning to walking would cause the pawn to "skip"
		//because falling and walking collision is calculated differently
		if(NextStateName=='PlayerWalking')
		{
			if(advPawn.wallClimbingState != EWCS_JumpBack)
			{
				Pawn.SetPhysics(PHYS_Walking);
			}
		}
	}

	event NotifyOnAnimEnd(AnimNodeSequence SeqNode, float PlayedTime, float ExcessTime)
	{
		switch(advPawn.wallClimbingState)
		{
		case EWCS_ToCeiling:
			Pawn.bCollideWorld = true;
			GotoState('PlayerCeilingClimbing');
			break;
		case EWCS_ScaleUp:
			Pawn.bCollideWorld = true;
			GotoState('PlayerWalking');
			break;
		case EWCS_JumpBack:
			GotoState('PlayerWalking');
			break;
		default:
			Pawn.bCollideWorld = true;
			RecalculateClimbingPosition();
			advPawn.wallClimbingState = EWCS_Default;
		}

		advPawn.DeactivateRootMotion();
	}

	function Vector GetDirectionalizedInputMovement(float _fDeltaTime)
	{
		local Vector	X,Y,Z,vInputVector;

		vInputVector = GetRelativeInputMovement(_fDeltaTime);

		//Prevent speed hack (e.g. Forward is slower than Foward + Right movement)
		if(VSize(vInputVector)>1)
			vInputVector = Normal(vInputVector);

		//Align acceleration to view angle
		GetAxes(Pawn.Rotation,X,Y,Z);

		vInputVector = Normal(vInputVector.X*vect(0,0,1) + (vInputVector.Y*vect(0,1,0)));

		return vInputVector;
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector	vNewAccel, vHitLocation, vHitNormal;
		local Vector    vGlobalAccel, vDir;
		local Vector vStart, vEnd;
		local Actor newBase;
		local AdvKitPhysicalMaterialProperty matProp;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			UpdateRotation( _fDeltaTime );
			if(!advPawn.bInRootMotion && !advPawn.bInRootRotation)
			{
				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);
				vNewAccel*= (vect(0,0,1)*advPawn.vWallClimbSpeed.Z + vect(0,1,0)*advPawn.vWallClimbSpeed.Y)*_fDeltaTime;

				if(CheckForClimbLocationInDirection(vNewAccel))
				{
					vGlobalAccel.Y = vNewAccel.Y;
					vGlobalAccel.X = vNewAccel.Z;
					vGlobalAccel = vGlobalAccel >> Pawn.Rotation;
					Pawn.Acceleration = vGlobalAccel;


					vNewAccel = advPawn.Global2RelativeDirection(vNewAccel >> Pawn.Rotation);

					vDir = Normal(GetRelativeInputMovement(_fDeltaTime));
					advPawn.fAnimTreeFrontDirectionBlend = Lerp(advPawn.fAnimTreeFrontDirectionBlend, vect(1,0,0) dot vDir,_fDeltaTime*10);
					advPawn.fAnimTreeSideDirectionBlend = Lerp(advPawn.fAnimTreeSideDirectionBlend, vect(0,1,0) dot vDir,_fDeltaTime*10);

					Pawn.SetRelativeLocation(Pawn.RelativeLocation+vNewAccel);
					//NOTE: Use this instead of the above line to prevent the pawn from "skipping" when starting to climb (does not work if the pole is already a based actor)
					//Pawn.SetRelativeLocation(advPawn.Global2RelativePosition(Pawn.Location)+vNewAccel); 

					vStart=Pawn.Location;
					vEnd=vStart+(vect(1,0,0) >> Pawn.Rotation)*Pawn.GetCollisionRadius()*2;
					newBase = TraceAdvanced(vStart,vEnd,,false,vHitLocation,vHitNormal,,,matProp);

					if(matProp!=none && matProp.bIsClimbableWall)
					{
						if(Pawn.Base != newBase)
						{
							advPawn.SetCurrentWall(newBase,vHitNormal);
							RecalculateClimbingPosition();
						}
					}
				}

				if(vNewAccel.Z<0 && CheckForFloor() || PlayerInput.aUp<0)
				{
					GotoState('PlayerWalking');
				}

				//Jump, obviously
				if (bPressedJump)
				{
					Pawn.DoJump(bUpdating);
					bPressedJump = false;
				}
			}
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		Pawn.SetRelativeRotation(Rotator(-advPawn.vCurrentWallNormal));
	}
}

state PlayerCeilingClimbing
{
	event BeginState(name PreviousStateName)
	{
		if(advPawn==none)
		{
			return;
		}

		advPawn.ceilingClimbingState = ECCS_Default;

		Pawn.SetCollision(true,true,false);
		Pawn.bCollideWorld = true;

		Pawn.SetPhysics(PHYS_Interpolating);
		Pawn.SetRelativeLocation(advPawn.Global2RelativePosition(Pawn.Location));
		Pawn.SetRelativeRotation(Rotator(advPawn.Global2RelativeDirection((vect(1,0,0) >> Pawn.Rotation))));
		advPawn.adventureState = EAS_CeilingClimbing;
	}

	event EndState(name NextStateName)
	{
		if(NextStateName!='PlayerWallClimbing')
		{
			Pawn.SetPhysics(PHYS_Falling);
		}
	}


	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.ceilingClimbingState)
		{
		case ECCS_ClimbToWall:
			GotoState('PlayerWallClimbing');
			break;
		}
		

		advPawn.DeactivateRootMotion();
	}


	function PlayerMove(float _fDeltaTime)
	{
		local Vector	vNewAccel;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion)
			{
				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);
				vNewAccel *= _fDeltaTime*advPawn.fCeilingClimbSpeed;
				
				Pawn.Acceleration = vect(0,0,0);
				if(CheckForCeilingLocationInDirection(vNewAccel))
				{
					Pawn.Acceleration = Normal(vNewAccel)*Pawn.AccelRate;
					Pawn.SetRelativeLocation(Pawn.RelativeLocation+(vNewAccel << Pawn.Base.Rotation));
				}

				//Check for wall (NotifyHitWall does not work if checking against Base)
				if(TraceForWallClimbSurface(Pawn.Location, Pawn.Location+Normal(vNewAccel)*advPawn.fWallTraceLength, advPawn.vCurrentWallNormal))
				{
					advPawn.ceilingClimbingState = ECCS_ClimbToWall;
					advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
				}

				//Jump, obviously
				if (bPressedJump  || PlayerInput.aUp<0)
				{
					Pawn.DoJump(bUpdating);
					bPressedJump = false;
					GotoState('PlayerWalking');
				}
			}

			// Update rotation.
			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		local Vector vTmpDirection;

		vTmpDirection = GetDirectionalizedInputMovement(_fDeltaTime);
		vTmpDirection.Z = 0;
		vTmpDirection = vTmpDirection << Pawn.Base.Rotation;
		if(VSize(vTmpDirection)!=0)
		{
			Pawn.SetRelativeRotation(RLerp(Pawn.RelativeRotation,Rotator(vTmpDirection),FMin(1,advPawn.fRotationLerpSpeed*_fDeltaTime),true));
		}
	}
}

state PlayerWallJumping
{
	//the pawn cannot be rotated while hanging on the wall
	ignores UpdatePawnRotation;

	event BeginState(name PreviousStateName)
	{
		if(advPawn==none)
		{
			//No time available, fall down instantly
			FallDown();
			return;
		}

		advPawn.adventureState = EAS_WallJumping;
		advPawn.wallJumpingState = EWJS_Default;
		Pawn.ZeroMovementVariables();
		Pawn.SetPhysics(PHYS_Interpolating);

		//start the fall down countdown
		SetTimer(advPawn.fWallJumpFallDownTime,false,'FallDown');
	}

	event EndState(Name NextStateName)
	{
		//cancel countdown in case player jumped
		advPawn.SetCurrentWall(none);
		ClearTimer('FallDown');
	}


	event NotifyOnAnimEnd(AnimNodeSequence SeqNode, float PlayedTime, float ExcessTime)
	{
		switch(advPawn.wallJumpingState)
		{
		case EWJS_JumpBack:
			GotoState('PlayerWalking');
			break;
		}

		advPawn.DeactivateRootMotion();
	}

	function PlayerMove(float _fDeltaTime)
	{
		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			UpdateRotation(_fDeltaTime);

			if (bPressedJump)
			{
				Pawn.DoJump(bUpdating);
				ClearTimer('FallDown');
				bPressedJump = false;
			}
		}
	}
}

state PlayerSplineMoving
{
	event BeginState(Name PreviousStateName)
	{
		Pawn.bCollideWorld=false;
		Pawn.ZeroMovementVariables();
		Pawn.SetPhysics(PHYS_Interpolating);
	}

	event EndState(Name NextStateName)
	{
		Pawn.bCollideWorld=true;
		advPawn.SetCurrentSpline(none);
	}
}

state PlayerLedgeClimbing extends PlayerSplineMoving
{
	event BeginState(Name PreviousStateName)
	{
		super.BeginState(PreviousStateName);
		advPawn.adventureState = EAS_LedgeClimbing;
	}

	event NotifyPreSplineEnd(optional AdvKitSplineActor _nextSplineActor, optional float _fDistToChange)
	{
		local Vector vNextLedgeNormal;
		local float fNextDistance;
		local float fAngle;

		if(_nextSplineActor!=none)
		{
			//calculate a distance on the next spline to determine normals
			if(advPawn.currentSpline.IsConnectedTo(_nextSplineActor,false))
			{
				fNextDistance = FMin(Pawn.GetCollisionRadius(),_nextSplineActor.GetSplineLength());
			}
			else if(_nextSplineActor.IsConnectedTo(advPawn.currentSpline,false))
			{
				fNextDistance = FMax(_nextSplineActor.GetSplineLength()-Pawn.GetCollisionRadius(),0);
			}

			//check whether the edge angle is pointed enough to do a climb around corner animation
			vNextLedgeNormal = _nextSplineActor.GetNormalAtDistanceAlongSpline(fNextDistance);
			vNextLedgeNormal = advPawn.Global2RelativeDirection(vNextLedgeNormal);

			fAngle = Atan2(vNextLedgeNormal.X,vNextLedgeNormal.Y)*RadToDeg;

			if(fAngle>=advPawn.fLedeAngleCornerLeft)
			{
			 	advPawn.SetCurrentSpline(_nextSplineActor,fNextDistance);
				advPawn.ledgeClimbingState = ELCS_ClimbAroundEdgeLeft;
				advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
			}
			if(fAngle<=advPawn.fLedeAngleCornerRight)
			{
			 	advPawn.SetCurrentSpline(_nextSplineActor,fNextDistance);
				advPawn.ledgeClimbingState = ELCS_ClimbAroundEdgeRight;
				advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
			}
		}
	}


	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.ledgeClimbingState)
		{
		case ELCS_FallingLedgeGrab:
			break;
		case ELCS_ClimbUp:
			Pawn.SetPhysics(PHYS_Walking);
			Pawn.bCollideWorld = true;
			GotoState('PlayerWalking');
			break;
		case ELCS_JumpDown:
			Pawn.SetPhysics(PHYS_Falling);
		case ELCS_JumpUp:
		case ELCS_JumpLeft:
		case ELCS_JumpRight:
			Pawn.SetCollision(true, true);
			Pawn.bCollideWorld = true;
		case ELCS_JumpBack:
			//Reset state to force blend node to switch animation
			advPawn.ledgeClimbingState = ELCS_Default; 
			GotoState('PlayerWalking');
			break;
		case ELCS_ClimbAroundEdgeLeft:
			break;
		case ELCS_ClimbAroundEdgeRight:
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector vTraceStart, vTraceEnd;
		local Vector vNewLocation;
		local Vector vNewAccel, vRelNewAccel;
		local float fDeltaDistance;
		local float fNormalAccelDotProduct;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion && !advPawn.bInRootRotation)
			{
				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);

				//update pawn's location
				vNewLocation = advPawn.vCurrentSplineRelativeLocation + advPawn.vCurrentSplineRelativeNormal*advPawn.vLedgeClimbOffset.X;
				Pawn.SetRelativeLocation(vNewLocation+advPawn.Global2RelativeDirection(vect(0,0,1)*advPawn.vLedgeClimbOffset.Z));
		
				//move depending on acceleration to tangent alignment
				vRelNewAccel = Normal(advPawn.Global2RelativeDirection(vNewAccel));
				fNormalAccelDotProduct = Normal(advPawn.vCurrentSplineRelativeNormal) dot vRelNewAccel;
				fDeltaDistance = (Normal(advPawn.vCurrentSplineRelativeTangent) dot vRelNewAccel)*_fDeltaTime;

				//check if there is a wall beneath the ledge
				advPawn.bLedgeClimbOnWall = CheckForWall();
				if(advPawn.bLedgeClimbOnWall)
				{
					//player is accelerating back, indicate back jump
					if(fNormalAccelDotProduct>0.9)
					{
						advPawn.ledgeClimbingState = ELCS_PreJumpBack;
					}
					//player is accelerating forward, indicate up jump or climb up
					else if(fNormalAccelDotProduct<-0.9)
					{
						vTraceStart = Pawn.Location + vect(0,0,1)*advPawn.fLedgeClimb_ClimbUpTraceHeight;
						vTraceEnd = vTraceStart + ((vect(1,0,0)*advPawn.fLedgeClimb_ClimbUpTraceLength) >> Pawn.Rotation);

						//the pawn can climb up
						if(FastTrace(vTraceEnd,vTraceStart))
						{
							advPawn.ledgeClimbingState = ELCS_ClimbUp;
							advPawn.ActivateRootMotion(RMM_Relative);
							return;
						}
						//indicate upwards jump
						else
						{
							advPawn.ledgeClimbingState = ELCS_PreJumpUp;
						}
					}
					else
					{
						//there is a wall beneath the ledge
						fDeltaDistance *= advPawn.fLedgeClimb_SpeedWall;
						advPawn.ledgeClimbingState = ELCS_Default;
					}
				}
				else
				{
					//there is no wall beneath the ledge, so the pawn cannot jump up 
					//or backwards because of missing wall support
					fDeltaDistance *= advPawn.fLedgeClimb_SpeedNoWall;
					advPawn.ledgeClimbingState = ELCS_Default;
				}


				//ledge is comming to an end so the pawn has to stop moving
				if(fDeltaDistance>0 && CheckSplineEnd(Pawn.GetCollisionRadius()+fDeltaDistance))
				{
					advPawn.SetSplineDistance(advPawn.currentSpline.GetSplineLength()-Pawn.GetCollisionRadius());
					vNewAccel = vect(0,0,0);
					fDeltaDistance=0;

					//in case there is a wall beneath the ledge the pawn can jump
					if(advPawn.bLedgeClimbOnWall)
					{
						if(advPawn.ledgeClimbingState != ELCS_PreJumpBack)
						{
							//do not attempt jump in case there is a wall to the side
							if(FastTrace(Pawn.Location+((vect(0,1,0)>>Pawn.Rotation)*(Pawn.GetCollisionRadius()+8)),Pawn.Location))
							{
								advPawn.ledgeClimbingState = ELCS_PreJumpRight;
							}
						}
					}
				}
				else if(fDeltaDistance<0 && CheckSplineEnd(-Pawn.GetCollisionRadius()+fDeltaDistance))
				{
					advPawn.SetSplineDistance(Pawn.GetCollisionRadius());
					vNewAccel = vect(0,0,0);
					fDeltaDistance=0;

					//in case there is a wall beneath the ledge the pawn can jump
					if(advPawn.bLedgeClimbOnWall)
					{
						if(advPawn.ledgeClimbingState != ELCS_PreJumpBack)
						{
							//do not attempt jump in case there is a wall to the side
							if(FastTrace(Pawn.Location+((vect(0,-1,0)>>Pawn.Rotation)*(Pawn.GetCollisionRadius()+8)),Pawn.Location))
							{
								advPawn.ledgeClimbingState = ELCS_PreJumpLeft;
							}
						}
					}
				}

				Pawn.Acceleration = vNewAccel*Pawn.AccelRate;//For animtree
				
				//update spline variables
				if(fDeltaDistance!=0)
				{
					advPawn.SetDeltaSplineDistance(fDeltaDistance);
				}

				if(bPressedJump)
				{
					bPressedJump = false;
					Pawn.DoJump(bUpdating);
				}

				//jump down from the ledge
				if(PlayerInput.aUp<0)
				{
					advPawn.ledgeClimbingState = ELCS_JumpDown;
					advPawn.ActivateRootMotion(RMM_Relative);
				}
			}

			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		local Vector tmpDirection;

		//align pawn to face the ledge
		tmpDirection = -advPawn.currentSpline.GetNormalAtDistanceAlongSpline(advPawn.fCurrentSplineDistance);
		if(VSize(tmpDirection)!=0)
		{
			Pawn.SetRotation(Rotator(tmpDirection));
		}
	}
}

state PlayerBalancing extends PlayerSplineMoving
{
	event BeginState(name PreviousStateName)
	{
		local AdvKitSplineActor tmpSpline;
		local float fNewDistance;
		local Vector vNewLocation;

		super.BeginState(PreviousStateName);

		if(advPawn==none)
		{
			return;
		}

		advPawn.adventureState = EAS_Balancing;

		//Reset balance values
		advPawn.fBalance = 0;
		advPawn.fTargetBalance = 0;

		//Find startspline
		if(!GetClosestSpline(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(), class'AdvKitSplineActorBalance', tmpSpline,fNewDistance,advPawn.fMaxBalanceCheckDistance,advPawn.fMaxBalanceAcceptDistance))
		{
			//None found, leave state
			GotoState('PlayerWalking');
			return;
		}

		advPawn.SetCurrentSpline(tmpSpline,fNewDistance);

		//The collision cylinder has an offset in PlayerWalking of 3 apparently, this compensates for it
		Pawn.SetCollisionSize(Pawn.GetCollisionRadius(),advPawn.GetCollisionHeightDefault()+3);

		vNewLocation = advPawn.vCurrentSplineRelativeLocation;
		vNewLocation.Z += Pawn.GetCollisionHeight();
		Pawn.SetRelativeLocation(Vlerp(Pawn.RelativeLocation,vNewLocation,0.1));
	}

	event EndState(Name NextStateName)
	{
		super.EndState(NextStateName);

		if(advPawn==none)
		{
			return;
		}

		//Reset values
		advPawn.DeactivateRootMotion();
		advPawn.SetCurrentSpline(none);
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector vYAxis;
		local Vector vRelNewAccel, vNewAccel;
		local Vector vNewLocation, vJumpOffDirection;
		local float fRandSteps, fDeltaBalance, fBalanceRestoreInput;
		local float fDeltaDistance;

		if(advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion)
			{
				vNewAccel = GetDirectionalizedInputMovement(_fDeltaTime);
				vRelNewAccel = Normal(advPawn.Global2RelativeDirection(vNewAccel));

				fDeltaDistance = (Normal(advPawn.vCurrentSplineRelativeTangent) dot vRelNewAccel)*advPawn.fBalancingWalkSpeed*_fDeltaTime;

				//Move to next spline actor if necessary
				if(fDeltaDistance!=0 && CheckSplineEnd(fDeltaDistance))
				{
					//Leave state in case spline path is finished
					GotoState('PlayerWalking');
					return;
				}
				advPawn.SetDeltaSplineDistance(fDeltaDistance);

				//Set new pawn location on spline
				vNewLocation = advPawn.vCurrentSplineRelativeLocation;
				vNewLocation.Z += Pawn.GetCollisionHeight();
				Pawn.SetRelativeLocation(Vlerp(Pawn.RelativeLocation,vNewLocation,0.1));

				//Control the animation with the AnimTree
				Pawn.Acceleration = vNewAccel*Pawn.AccelRate;

				//Handle balancing
				fRandSteps = 40;

				//Random loss of balance
				fDeltaBalance = ((fRandSteps-Rand(fRandSteps*2+1))/fRandSteps);
		
				//Input influences balance
				vYAxis = vect(0,1,0) >> Pawn.Rotation;
				fBalanceRestoreInput = (vYAxis dot vNewAccel);
				fDeltaBalance += fBalanceRestoreInput;

				//Not restoring balance will make balance loss increase over time
				if(Abs(fBalanceRestoreInput)<0.1)
				{
					fDeltaBalance += advPawn.fTargetBalance*0.6;
				}

				//Add delta balance to actual balance value
				advPawn.fTargetBalance += fDeltaBalance * advPawn.fBalanceDifficulty * _fDeltaTime;
				//Prevent value from getting too small/high
				advPawn.fTargetBalance = FClamp(advPawn.fTargetBalance,-1.1,1.1);

				//Set pawn's balance value for AnimTree
				//Lerp to keep it jitter-free
				advPawn.fBalance = Lerp(advPawn.fBalance,advPawn.fTargetBalance,0.1);

				//the player was not able to hold the balance, fall off
				if(Abs(advPawn.fBalance)>=1)
				{
					//jump off to the side the pawn is leaning to
					vJumpOffDirection = vect(0,1,0)*advPawn.fBalance;
					vJumpOffDirection.Z = 1;
					vJumpOffDirection *= advPawn.vBalanceFallOffSpeed;
					vJumpOffDirection = vJumpOffDirection >> Pawn.Rotation;

					Pawn.SetPhysics(PHYS_Falling);
					Pawn.Velocity = vJumpOffDirection;
					vJumpOffDirection.Z = 0;
					Pawn.SetRotation(Rotator(vJumpOffDirection));

					GotoState('PlayerWalking');
					return;
				}

				if(bPressedJump)
				{
					if(Pawn.DoJump(bUpdating))
					{
						GotoState('PlayerWalking');
					}
					bPressedJump = false;
				}
			}
		}

		// Update rotation.
		UpdateRotation( _fDeltaTime );
	}


	function UpdatePawnRotation(float _fDeltaTime)
	{
		local Rotator   rDestRotation;
		local Vector    vTmpDirection;
		local float     fDirectionDotProduct;

		//make the pawn face the direction that it is moving in
		if(VSize(Pawn.Acceleration)!=0)
		{
			vTmpDirection = advPawn.vCurrentSplineTangent;
			vTmpDirection.Z = 0;
			fDirectionDotProduct = Normal(vTmpDirection) dot Normal(Pawn.Acceleration);

			//some threashold to prevent unnecessary rotation flips
			if(Abs(fDirectionDotProduct)>0.2)
			{
				vTmpDirection *= fDirectionDotProduct;
				rDestRotation = Rotator(advPawn.Global2RelativeDirection(vTmpDirection));

				Pawn.SetRelativeRotation(RLerp(Pawn.RelativeRotation,rDestRotation,FMin(1,advPawn.fRotationLerpSpeed*_fDeltaTime),true));
			}
		}
	}
}

state PlayerTightSpaceWalking extends PlayerSplineMoving
{
	event BeginState(Name PreviousStateName)
	{
		local Vector vStartLocation;
		local AdvKitSplineActor tmpTightSpline;
		local float fNewDistance;

		super.BeginState(PreviousStateName);

		if(advPawn==none)
		{
			return;
		}

		advPawn.adventureState = EAS_TightSpaceWalking;

		if(!GetClosestSpline(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(), class'AdvKitSplineActorTightSpace', tmpTightSpline,fNewDistance,advPawn.fMaxTightspaceCheckDistance,advPawn.fMaxTightspaceAcceptDistance))
		{
			GotoState('PlayerWalking');
			return;
		}

		advPawn.SetCurrentSpline(tmpTightSpline,fNewDistance);

		//The collision cylinder has an offset in PlayerWalking of 3 apparently, this compensates for it
		Pawn.SetCollisionSize(Pawn.GetCollisionRadius(),advPawn.GetCollisionHeightDefault()+3);

		vStartLocation = advPawn.vCurrentSplineRelativeLocation + vect(0,0,1)*advPawn.vTightSpaceStartOffset.Z - Normal(advPawn.vCurrentSplineRelativeNormal)*advPawn.vTightSpaceStartOffset.Y;
		if(advPawn.fCurrentSplineDistance < advPawn.currentSpline.GetSplineLength()/2.0)
		{
			advPawn.tightSpaceState = ETSS_Enter_Wall_Right;
			Pawn.SetRelativeRotation(Rotator(advPawn.vCurrentSplineRelativeTangent));
			vStartLocation -= Normal(advPawn.vCurrentSplineRelativeTangent)*advPawn.vTightSpaceStartOffset.X;
		}
		else
		{
			advPawn.tightSpaceState = ETSS_Enter_Wall_Left;
			Pawn.SetRelativeRotation(Rotator(-advPawn.vCurrentSplineRelativeTangent));
			vStartLocation += Normal(advPawn.vCurrentSplineRelativeTangent)*advPawn.vTightSpaceStartOffset.X;
		}

		//All the debug lines show correct locations
		Pawn.SetRelativeLocation(vStartLocation);
		advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
	}

	event EndState(Name NextStateName)
	{
		super.EndState(NextStateName);

		if(advPawn==none)
		{
			return;
		}

		advPawn.SetCurrentSpline(none);
		//reset collision size
		Pawn.SetCollisionSize(Pawn.GetCollisionRadius(),advPawn.GetCollisionHeightDefault());
	}

	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		local float fNewDistance;

		switch(advPawn.tightSpaceState)
		{
		//finished entering animation, stick to spline
		case ETSS_Enter_Wall_Right:
			advPawn.tightSpaceState = ETSS_Default_Wall_Right;
			advPawn.currentSpline.GetClosestPositionOnSplineTo(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(),,fNewDistance);
			advPawn.SetSplineDistance(fNewDistance);
			advPawn.DeactivateRootMotion();
			break;
		//finished entering animation, stick to spline
		case ETSS_Enter_Wall_Left:
			advPawn.tightSpaceState = ETSS_Default_Wall_Left;
			advPawn.currentSpline.GetClosestPositionOnSplineTo(Pawn.Location-vect(0,0,1)*Pawn.GetCollisionHeight(),,fNewDistance);
			advPawn.SetSplineDistance(fNewDistance);
			advPawn.DeactivateRootMotion();
			break;
		case ETSS_Leave_Wall_Right:
		case ETSS_Leave_Wall_Left:
			GotoState('PlayerWalking');
			advPawn.DeactivateRootMotion();
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		local Vector vNewLocation;
		local Vector vInputVector;
		local Vector vTangent;
		local float fDeltaProgress;

		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			if(!advPawn.bInRootMotion)
			{
				vTangent = Normal(advPawn.currentSpline.GetTangentAtDistanceAlongSpline(advPawn.fCurrentSplineDistance));

				vInputVector = GetDirectionalizedInputMovement(_fDeltaTime);
				vInputVector.Z=0;
				vInputVector *= Abs(vTangent dot vInputVector);

				//scale pawn's acceleration depending on input alignment to spline's tangent
				Pawn.Acceleration = vInputVector*Pawn.AccelRate;
				fDeltaProgress = (Normal(advPawn.vCurrentSplineRelativeTangent) dot Normal(advPawn.Global2RelativeDirection(vInputVector)))*advPawn.fTightSpaceWalkSpeed*_fDeltaTime;

				//in case the spline is at an end, leave the tight space
				if(fDeltaProgress>0 && CheckSplineEnd(Pawn.GetCollisionRadius()))
				{
					advPawn.tightSpaceState = ETSS_Leave_Wall_Right;
					advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
					return;
				}

				//in case the spline is at an end, leave the tight space
				if(fDeltaProgress<0 && CheckSplineEnd(-Pawn.GetCollisionRadius()))
				{
					advPawn.tightSpaceState = ETSS_Leave_Wall_Left;
					advPawn.ActivateRootMotion(RMM_Relative,RMRM_RotateActor);
					return;
				}


				//update spline variables
				advPawn.SetDeltaSplineDistance(fDeltaProgress);

				//update paw's location
				vNewLocation = advPawn.vCurrentSplineRelativeLocation;
				vNewLocation.Z += Pawn.GetCollisionHeight();
				Pawn.SetRelativeLocation(VLerp(Pawn.RelativeLocation,vNewLocation,0.3));
			}

			bPressedJump = false;

			// Update rotation.
			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdatePawnRotation(float _fDeltaTime)
	{
		//align pawn to spline tangent
		Pawn.SetRelativeRotation(Rotator(-advPawn.vCurrentSplineRelativeNormal));
	}
}

state PlayerInteracting
{
	event BeginState(Name PreviousStateName)
	{
		Pawn.ZeroMovementVariables();
		if(advPawn!=none)
		{
			advPawn.adventureState = EAS_Interacting;
		}
	}

	event NotifyOnAnimEnd(AnimNodeSequence _SeqNode, float _fPlayedTime, float _fExcessTime)
	{
		switch(advPawn.interactionState)
		{
		//interaction is done, return to walking
		case EIS_ActivateLever:
		case EIS_DeactivateLever:
			Pawn.SetPhysics(PHYS_Walking);
			GotoState('PlayerWalking');
			advPawn.interactionState = EIS_None;
			break;
		}
	}

	function PlayerMove(float _fDeltaTime)
	{
		if (advPawn == none)
		{
			GotoState('Dead');
		}
		else
		{
			//Update rotation, nothing more
			//only the camera may rotate via player input
			UpdateRotation( _fDeltaTime );
		}
	}

	function UpdateRotation( float _fDeltaTime )
	{
		local Rotator	rDeltaRot, rViewRotation;

		//Manage and rotate the camera
		rViewRotation = Rotation;

		// Calculate Delta to be applied on ViewRotation
		rDeltaRot.Yaw	= PlayerInput.aTurn;
		rDeltaRot.Pitch	= PlayerInput.aLookUp;

		ProcessViewRotation( _fDeltaTime, rViewRotation, rDeltaRot );
		SetRotation(rViewRotation);

		ViewShake( _fDeltaTime );
	}
}

DefaultProperties
{
	CameraClass = class 'AdvKit.AdvKitCamera'
	MinHitWall = 0.8
	fMinCameraDistance = 64
	fMaxCameraDistance = 512
}