/**
 * The main extended Spline Actor used in the adventure kit
 * 
 * 2013 by FreetimeCoder
 * www.freetimestudio.net
 */
class AdvKitSplineActor extends SplineActor
	abstract
	ClassGroup(AdvKit);

/**the next actor to activate the connection to for a bridge*/
var(Bridge) AdvKitSplineActor nextActor;
/**the previous actor to activate the connection to for a bridge*/
var(Bridge) AdvKitSplineActor prevActor;

/**is the bridge active or not*/
var(Bridge) bool bBridgeActive;

/**
 * storing the last location to only update
 * spline components when they change
 */
var Vector  vLastLocation;
/**
 * storing the last rotation to only update
 * spline components when they change
 */
var Rotator rLastRotation;

/**Set initial variables*/
function PostBeginPlay()
{
	vLastLocation = Location;
	rLastRotation = Rotation;
}

/**Called every frame to update
 * 
 * @param   _fDeltaTime     the time passed since last call
 */
event Tick(float _fDeltaTime)
{
	//only update if the actor is movable
	if(bMovable)
	{
		//only update if the actor's transformation changed
		if(vLastLocation!=Location || rLastRotation != Rotation)
		{
			//update components
			UpdateSplineComponents();
		}
	}
}

/**Query if this actor has a next connection
 * 
 * @return  if there is a next connection
 */
final function bool HasNext()
{
	//return bridge or connection size
	return (bBridgeActive && nextActor!=none) || (Connections.Length>0);
}

/**Query if this actor has a previous connection
 * 
 * @return  if there is a next connection
 */
final function bool HasPrevious()
{
	//return bridge or linksfrom size
	return (bBridgeActive && prevActor!=none) || (LinksFrom.Length>0);
}

/**Get the next AdvKitSplineActor on the curve
 * 
 * @return  the next spline
 */
function AdvKitSplineActor GetNext()
{
	//bridge connection is primary
	if(bBridgeActive && nextActor!=none)
	{
		return nextActor;
	}

	//return the first connection if there are any
	if(Connections.Length>0)
	{
		return AdvKitSplineActor(Connections[0].ConnectTo);
	}

	//there is no next
	return none;
}

/**Get the previous AdvKitSplineActor on the curve
 * 
 * @return  the previous spline
 */
function AdvKitSplineActor GetPrevious()
{
	//bridge connection is primary
	if(bBridgeActive && prevActor!=none)
	{
		return prevActor.GetPrevious();
	}

	//return the first connection if there are any
	if(LinksFrom.Length>0)
	{
		return AdvKitSplineActor(LinksFrom[0]);
	}

	//there is no next
	return none;
}

/**Get the length of the first connection from this to the next actor
 * 
 * @return  the length
 */
final function float GetSplineLength()
{
	//make sure there is a connection
	if(Connections.Length>0)
	{
		return Connections[0].SplineComponent.GetSplineLength();
	}

	//no connection
	return 0;
}

/**Calculate the tangent of the spline curve at a given location on the spline
 * This function wraps the tangent function for the first spline component
 * 
 * @param   _fDistance  the location on the curve
 * @return              the tangent
 */
final function Vector GetTangentAtDistanceAlongSpline(float _fDistance)
{
	if(Connections.Length>0)
	{
		return Connections[0].SplineComponent.GetTangentAtDistanceAlongSpline(_fDistance);
	}

	return SplineActorTangent;
}

/**Calculate the normal of the spline curve at a given location on the spline
 * 
 * @param   _fDistance  the location on the curve
 * @return              the normal
 */
final function Vector GetNormalAtDistanceAlongSpline(float _fDistance)
{
	local Vector vNormal;
	vNormal = GetTangentAtDistanceAlongSpline(_fDistance);
	vNormal = Normal(vNormal cross vect(0,0,-1));
	return vNormal;
}

/**Calculate the location of the spline curve at a given location on the spline
 * This function wraps the location function for the first spline component
 * 
 * @param   _fDistance  the location on the curve
 * @return              the 3d location in global space
 */
final function Vector GetLocationAtDistanceAlongSpline(float _fDistance)
{
	if(Connections.Length>0)
	{
		return Connections[0].SplineComponent.GetLocationAtDistanceAlongSpline(_fDistance);
	}

	return Location;
}

/**Find the closest location on the spline curve to a given location in global space
 * using binary search
 * 
 * @param   _vRefPosition   the location in global space
 * @param   _vOutPosition   the found location on the spline
 * @param   _fOutLength     the distance on the spline of the found location
 * @param   _fEpsilon       how precise the result should be
 * @param   _iMaxIterations how many iterations to allow before to break even if targeted precision was not reached
 * @return                  the distance to the found location
 */
final function float GetClosestPositionOnSplineTo(Vector _vRefPosition, optional out Vector _vOutPosition, optional out float _fOutLength, optional float _fEpsilon, optional int _iMaxIterations)
{
	local float fLengthLeft, fLengthMid, fLengthRight;
	local Vector vLocLeft, vLocMid, vLocRight;
	local float fDistLeft, fDistMid, fDistRight;
	local float fLastDistance, fSplineLength;
	local float fCurrentEps;
	local int iCurrentIteration;

	if(_iMaxIterations<=0)
	{
		_iMaxIterations = 50;
	}

	fSplineLength = GetSplineLength();

	fLengthLeft = 0;
	fLengthRight = fSplineLength;

	fLastDistance = 1000000;
	iCurrentIteration = 0;

	//Basically this is just a simple binary search with left, mid and right value
	vLocLeft = GetLocationAtDistanceAlongSpline(fLengthLeft);
	vLocRight = GetLocationAtDistanceAlongSpline(fLengthRight);
	fDistLeft = VSize(vLocLeft - _vRefPosition);
	fDistRight = VSize(vLocRight - _vRefPosition);

	do
	{
		fLengthMid = (fLengthLeft + fLengthRight)*0.5;
		vLocMid = GetLocationAtDistanceAlongSpline(fLengthMid);
		fDistMid = VSize(vLocMid - _vRefPosition);

		if(fDistLeft < fDistRight)
		{
			fLengthRight = fLengthMid;
			vLocRight = vLocMid;
			fDistRight = fDistMid;
		}
		else
		{
			fLengthLeft = fLengthMid;
			vLocLeft = vLocMid;
			fDistLeft= fDistMid;
		}

		fCurrentEps = Abs(fLastDistance - fDistMid);
		fLastDistance = fDistMid;
		iCurrentIteration++;
	}
	//abort when the delta distance is smaller than epsilon, too many iterations have been made, both borders are the same or the point is either of the splines ends
	until(fCurrentEps<=_fEpsilon || iCurrentIteration>=_iMaxIterations || fDistLeft==fDistRight || fDistMid==0 || fDistMid==fSplineLength);

	_vOutPosition = vLocMid;
	_fOutLength = fLengthMid;

	return fDistMid;
}

DefaultProperties
{
	Begin Object Name=Sprite
		Sprite=Texture2D'EditorResources.Spline.T_Loft_Spline'
		Scale = 0.1
	End Object

	Physics=PHYS_Interpolating

	bNoDelete=true
	bEdShouldSnap=true
	bMovable = true
	bStatic = false
}
